// irreducibility testing of finite nilpotent matrix groups

declare verbose Irrednil, 1;

import "fgrp.m": FGrp, FElt, FDot, MultFElt, InverseFElt, GroupifyFElt,
    GrpToFGrp, IsAbelianFGrp, DoCommuteFElt, SylowSubgroupFSub,
    ExponentElementFSub;
import "anc.m": IsIrreducibleANC;
import "decomp.m": ABELIAN_FULL, ABELIAN_SEMI, ABELIAN_DECIDE,
    HomogeneousDecompositionAbelian;
import "noncyclic.m": HasNoncyclicAbelian;
import "congruence.m": CongruenceObject, Modp,
    CongruencePrime, CongruenceField;
import "misc.m": MyIsFormallyReal, IndexProperty, InducedMatrixGroup,
    CanHandleField, MinimalPolynomialSemisimpleMatrix, PandPPrimePart;

/*
 * In the paper, the algorithm is described for group acting on
 * an abstract vector space V. The reduction technique (the
 * function `Reduce' in the paper) can replace V by a subspace U.
 * In this setting, if a vector which generates a proper submodule is
 * found, then it will be *contained* in the original space V.
 * In the implementation, all the spaces are of the form V=K^d, whence
 * we need to keep track of the base changes underlying a transition
 * from V to U. For this purpose, many functions below take an extra
 * argument `mu' which should be a matrix which is applied to subspaces
 * returned by functions.
 */
 
function IsIrreducibleAbelian( G : mu := false, DecideOnly := false, RandomThreshold := -1 )
    vprint Irrednil: "IsIrreducibleAbelian";

    if Degree(G) le 1 then
	return true, _;
    end if;
    
    apply_mu := func<x | mu cmpeq false select x else x * mu>;

    if not DecideOnly then
       /*
        * If G is inhomogeneous, return one of the homogeneous components.
        */
        is_homg, M := HomogeneousDecompositionAbelian(G : mode := ABELIAN_SEMI, RandomThreshold := RandomThreshold);
        if not is_homg then
            return false, apply_mu(M.1);
            
        end if;
    else
        is_homg := HomogeneousDecompositionAbelian(G : mode := ABELIAN_DECIDE, RandomThreshold := RandomThreshold);
        if not is_homg then
            return false, _;
        end if;
    end if;    

   /*
    * If G is homogeneous, then it is irreducible iff any non-zero
    * vector generates a proper submodule.
    */
    VG := GModule(G);
    V := RSpace(G);

    UG := sub<VG | V.1>;
    U := sub<V | [V!VG!u : u in Basis(UG)]>;

    if Dimension(U) eq Dimension(V) then
	return true, _;
    elif not DecideOnly then
	return false, apply_mu(V.1);
    else
        return false, _;
    end if;
end function;

/*
 * ReduceDimension
 *
 * `res' will either be the mu-image of a proper submodule
 * for G, res = false (submodule exists but not constructed)
 *  or res = true, signifying that G and mu have been updated
 */
procedure ReduceDimension(~G, F, ~mu, decomp, ~res : DecideOnly := false)
    vprint Irrednil: "ReduceDimension";

    /*
    * Find permutation rep for the action of G on decomp
    */
    //gens := [G.i : i in [1..Ngens(G)]];
    gens := [FDot(F,i) : i in [1..F`ngens]];
    m := #gens;
    n := #decomp;

    orb := [1];
    tra := [FDot(F,0)]; //Id(G)
    sta := [];
    
    vprintf Irrednil:
        "\t orbit stabiliser computations on %o points... ", n;
    i := 0;
    while i lt #orb do
        i +:= 1;
        for j in [1 .. m] do
           /*
            * We only need to test membership of ONE basis vector.
            */
            //w := Index(decomp, decomp[orb[i]] * G.j);
            w := IndexProperty(decomp, func<U| vec in U>)
                where vec is (decomp[orb[i]].1) * G.j;
            k := Index( orb, w );
            if k eq 0 then
                Append( ~orb, w );
                Append( ~tra, MultFElt(tra[i], FDot(F,j)));
            else
                s := MultFElt(MultFElt(tra[i],FDot(F,j)),
                    InverseFElt(tra[k]));
                pred := func<x|x`elt eq s`elt>;
                if IndexProperty(sta,pred) eq 0 then
                    Append(~sta, s);
                end if;
            end if;
        end for;
    end while;
    vprint Irrednil: "done";
    
    apply_mu := func<x | mu cmpeq false select x else x * mu>;
    if #orb lt n then
       /*
        * Action is intransitive.
        */
        vprintf Irrednil:
            "\t action is intransitive (#orbit = %o)\n", #orb;
        res := DecideOnly select false else apply_mu(decomp[1].1);
        return;
    else
       /*
        * Transitive action: Replace G and mu.
        */
        vprint Irrednil: "\t action is transitive";
        
        vprintf Irrednil: "\t lifting stabiliser... ";
        sta := [GroupifyFElt(G,h) : h in sta];
        vprint Irrednil: "done";
        vprintf Irrednil:
        "\t generating new matrix group in degree %o on %o generators... ",
        Dimension(decomp[1]), #sta;
        
        G, lambda := InducedMatrixGroup( sta, decomp[1] );
        vprint Irrednil: "done";
        mu := lambda * mu;
        res := true;
        return;
    end if;
end procedure;

function IsIrreducibleNilpotentRelative
             (G, mu : DecideOnly := false, CheapTests := false,
                   // MeatAxe := true,
                      RandomThreshold := -1,
                      NeqnThreshold := -1)
    local res;

    vprint Irrednil: "IsIrreducibleNilpotentRelative";

    // outer loop: group changed because of reduction
    // at termination:
    // if an inhomgeneous subgroup was found, then bo = false
    // and decomp is the homg decomp
    // otherwise: bo = true, 
    repeat
        vprint Irrednil: "Current degree:", Degree(G);
        if Degree(G) eq 1 then
            return true, _;
        end if;
        vprint Irrednil: "Number of generators:", Ngens(G);
        
        if CheapTests and DecideOnly and IsOdd(Degree(G))
                 and MyIsFormallyReal(BaseRing(G)) then
               /*
                * If DecideOnly = false, we should be able to find a
                * submodule easily beause of the following:
                *   If G is finite nilpotent over a formally real field,
                *   and if G is homogeneous then G is scalar or
                *   Degree(G) is even. (Easy!)
                */
                return false, _;
        end if;
        
       /*
        if MeatAxe then
            r, x := IsIrreducibleMeatAxe(G : mu := mu);
            if Type(r) eq BoolElt then
                vprint Irrednil: "\t the Meat-Axe did it";
                if DecideOnly then
                    return r, _;
                elif assigned x then
                    return r, x;
                else
                    return r, _;
                end if;
            else
                vprint Irrednil: "\t the Meat-Axe failed";
            end if;
        end if;
        */

        vprintf Irrednil: "\t getting congruence homomorphism... ";
        p := CongruenceObject(G);
        vprintf Irrednil: "codomain: GF(%o^%o)\n", 
            q, Valuation(#CongruenceField(p),q)
            where q is CongruencePrime(p);
        
        vprintf Irrednil: "\t setting up congruent image... ";
        if CheapTests then
           /*
            * Computing the congruent image requires virtually no extra
            * work in addition to Modp'ing the G.i.
            * We might then as well use the Meat-Axe for this.
            */
            Gmodp := Modp(G,p);
            if IsIrreducible(Gmodp) then
                vprint Irrednil: "aborted";
                vprint Irrednil: "\t congruent image is irreducible";
                return true, _;
            end if;
            F := GrpToFGrp(Gmodp);
        else
            gens := [Modp(G.i,p) : i in [1..Ngens(G)]];
           /*
            * Inversion mod p seems to be faster than Modp(G.-i,p).
            * Not too surprising... at least it's polynomial time!
            */
            inv  := [x^(-1) : x in gens];
            F := rec<FGrp | ngens := Ngens(G),
                            id := IdentityMatrix(CongruenceField(p),
                                                 Degree(G)),
                        gens := gens, inv := inv>;
        end if;
        vprintf Irrednil: "done\n";
        
        vprintf Irrednil: "\t is the group abelian? ";
	if IsAbelianFGrp(F) then
            vprint Irrednil: "yes";
	    return IsIrreducibleAbelian
                       (G : mu := mu, DecideOnly := DecideOnly, RandomThreshold := RandomThreshold);
	end if;
        vprint Irrednil: "no";

        if Degree(G) eq 2 then
            vprint Irrednil:
                "\t a non-abelian group of degree 2 is irreducible";
            return true, _;
        end if;

        r, A, H, H2_is_Q8 := HasNoncyclicAbelian(F);
        if r then
            vprint Irrednil:
                "\t found non-cyclic abelian normal subgroup";
            vprintf Irrednil:
                "\t lifting subgroup... ";
            // Can we get rid of this as well? We don't necessarily
            // need *all* the inverses of generators of A.
            A := sub<G|[GroupifyFElt(G,a) : a in A]>;
            vprint Irrednil: "done";
            decomp := HomogeneousDecompositionAbelian(A : RandomThreshold := RandomThreshold);
        else
            vprint Irrednil: "\t ANC group detected";

           /*
            * Find cyclic H < G of index 2.
            * Unless `H2_is_Q8', this has already been done by
            * `HasNoncyclicAbelian'
            */
            vprintf Irrednil: "\t finding cyclic maximal subgroup ";
            if H2_is_Q8 then
                // take Hall 2'-subgroup and any non-central element
                F2prime := SylowSubgroupFSub(F, -2);
                _ := exists(g){ g : i in [1..F`ngens] |
                    exists{j : j in [1..F`ngens] |
                    not DoCommuteFElt(g, FDot(F,j))}
                     where g is FDot(F,i)};
                H := F2prime cat [g];
            end if;
            vprint Irrednil: "done";

           /*
            * Find a generator of H and make sure that it is homogeneous.
            */
            h := ExponentElementFSub(H);
            vprintf Irrednil:
                "\t lifting generator (#bio = %o)... ", #h`bio;
            x := GroupifyFElt(G,h);
            vprint Irrednil: "done";
            
            vprintf Irrednil: "\t is it homogeneous? ";

            f := Factorisation(MinimalPolynomialSemisimpleMatrix(x));
            if #f gt 1 then
                vprint Irrednil: "no";
                decomp := [ Kernel(Evaluate(g[1], A!x)) : g in f  ]
                    where A is MatrixAlgebra(BaseRing(x), Nrows(x));
                A := H;
            else
                vprint Irrednil: "yes";
                
                vprintf Irrednil: "\t setting up nice generators... ";
                xp := h`elt;
                _ := exists(i){i : i in [1..Ngens(G)]
                                 | not DoCommuteFElt(FDot(F,i),h)};
                g := G.i;
                gp := FDot(F,i)`elt;

                _, pp := PandPPrimePart( Order(gp), 2);
                g := g^pp;
                gp := gp^pp;
                
                order_G := 2 * Order(xp);

                _, pp := PandPPrimePart(Order(xp), 2);
                xp := xp^pp;
                
                I := (gp)^0; 
                if (xp * xp^gp eq -I) and (gp^2 ne I) then
                    g := x^pp * g;
                    gp := Modp(g,p);
                end if;
                eps := (gp^2 eq I) select 1 else -1;
                vprint Irrednil: "done";
                
                // G := MatrixGroup< Degree(G), BaseRing(G) | g,x>;
                r, x := IsIrreducibleANC(g, x, order_G, eps :
                            mu := mu, DecideOnly := DecideOnly,
                            NeqnThreshold := NeqnThreshold);
                break;
            end if;
        end if;

       /*
        * A is definitely inhomogeneous at this stage and `decomp' is
        * the homogeneous decomposition of its natural module.
        */
	ReduceDimension(~G, F, ~mu, decomp, ~res :
            DecideOnly := DecideOnly);
	if res cmpeq true then
            vprint Irrednil: "\t reduction performed, starting again";
	    continue; // start again
	else
            r := false; // reducible action; res generates a submodule
                        // or res = false
	    x := res;
	    break;
	end if;
    until false;

   /*
    * At this point `r' signifies irreducibility of G and x generates
    * a proper submodule of the original group. In other words, mu
    * has already been applied. This allows us to pass on control
    * to other functions (*Abelian, *ANC, *Meataxe)
    */
    if r then
	return true, _;
    elif (not r) and (DecideOnly or (not assigned x)) then
	return false, _;
    else
	return false, x;
    end if;
end function;

intrinsic IsIrreducibleFiniteNilpotent( G::GrpMat :
        DecideOnly      := false,
        VectorMode      := false,
        CheapTests      := false,
     // MeatAxe         := false,
        NeqnThreshold   := Infinity()
    ) -> BoolElt, .
    
{
  Test irreducibility of the finite nilpotent matrix group G defined
  over the rationals, a number field, or a rational function field over
  a number field. 
}
    require CanHandleField(BaseRing(G)):
    "Cannot handle groups over this base ring.";
    
    t0:= Cputime();
   /*
    * The important cases of the trivial group.
    */
    if Ngens(G) eq 0 then
        case Degree(G):
        when 0:
            return false, _;
        when 1:
            return true, _;
        else:
            return false, sub<V|V.1> where V is RSpace(G);
        end case;
        return Degree(G) eq 1, _;
    end if;
    
    
    verify := func< W | forall{g : g in Generators(G) | W * g eq W }>;
	
    r, x := IsIrreducibleNilpotentRelative
              (G, IdentityMatrix(BaseRing(G),Degree(G)) :
              DecideOnly := DecideOnly, CheapTests := CheapTests,
           // MeatAxe := false,
              NeqnThreshold := NeqnThreshold,
              RandomThreshold := -1 // IOW: don't ever use the randomised
                                    // method; hard to debug and not even
                                    // much faster when it's fast (which
                                    // is not as often as one might think)
              );

    vprintf Irrednil: "\t TIME: vector-mode completed after %o\n",
        Cputime(t0);

    if r then
	return true, _;
    elif DecideOnly then
        return false, _;
    elif not assigned x then
        vprint Irrednil:
            "WARNING: Group was found to be reducible but no submodule \
has been constructed.";
	return false, _;
    elif VectorMode then
        return false, x;
    else
        vprintf Irrednil: "\t generating submodule... ";
        V := RSpace(G);
        MG := sub<GModule(G)|x>;
        vprint Irrednil: "done";
        //MG, f := sub<GModule(G)|x>;
        //U :=  VectorSpaceWithBasis([V ! (m @ f) : m in Basis(MG) ]);
        vprintf Irrednil: "\t TIME: all done after %o\n", Cputime(t0);
        return false, MG;
        //return false, U;
    end if;
end intrinsic;
